import math
from typing import (Optional,
                    Tuple)

from .hints import Scalar
from .point import Point
from .segment import Segment


def sign(first_point: Point, second_point: Point, third_point: Point) -> int:
    determinant = ((first_point.x - third_point.x)
                   * (second_point.y - third_point.y)
                   - (second_point.x - third_point.x)
                   * (first_point.y - third_point.y))
    if determinant > 0:
        return 1
    elif determinant < 0:
        return -1
    else:
        return 0


def _find_intersections(u0: Scalar, u1: Scalar, v0: Scalar, v1: Scalar
                        ) -> Tuple[int, Optional[Scalar], Optional[Scalar]]:
    first_coefficient = second_coefficient = None
    if u1 < v0 or u0 > v1:
        return 0, first_coefficient, second_coefficient
    if u1 > v0:
        if u0 < v1:
            first_coefficient = v0 if (u0 < v0) else u0
            second_coefficient = v1 if (u1 > v1) else u1
            return 2, first_coefficient, second_coefficient
        else:
            first_coefficient = u0
            return 1, first_coefficient, second_coefficient
    else:
        first_coefficient = u1
        return 1, first_coefficient, second_coefficient


def find_intersections(first_segment: Segment, second_segment: Segment,
                       *,
                       threshold: float = 1e-8,
                       squared_inv_epsilon: int = 10 ** 7
                       ) -> Tuple[int, Optional[Point], Optional[Point]]:
    first_source, second_source = first_segment.source, second_segment.source
    first_target, second_target = first_segment.target, second_segment.target
    first_vector = Point(first_target.x - first_source.x,
                         first_target.y - first_source.y)
    second_vector = Point(second_target.x - second_source.x,
                          second_target.y - second_source.y)
    e = Point(second_source.x - first_source.x,
              second_source.y - first_source.y)
    cross_product = (first_vector.x * second_vector.y
                     - first_vector.y * second_vector.x)
    squared_cross_product = cross_product * cross_product
    first_squared_length = (first_vector.x * first_vector.x
                            + first_vector.y * first_vector.y)
    second_squared_length = (second_vector.x * second_vector.x
                             + second_vector.y * second_vector.y)
    if squared_cross_product > (first_squared_length * second_squared_length
                                / squared_inv_epsilon):
        s = (e.x * second_vector.y - e.y * second_vector.x) / cross_product
        if (s < 0) or (s > 1):
            return 0, None, None
        t = (e.x * first_vector.y - e.y * first_vector.x) / cross_product
        if (t < 0) or (t > 1):
            return 0, None, None
        first_intersection_point = Point(first_source.x + s * first_vector.x,
                                         first_source.y + s * first_vector.y)
        if first_intersection_point.distance_to(first_source) < threshold:
            first_intersection_point = first_source
        elif first_intersection_point.distance_to(first_target) < threshold:
            first_intersection_point = first_target
        elif first_intersection_point.distance_to(second_source) < threshold:
            first_intersection_point = second_source
        elif first_intersection_point.distance_to(second_target) < threshold:
            first_intersection_point = second_target
        return 1, first_intersection_point, None
    e_squared_length = e.x * e.x + e.y * e.y
    cross_product = e.x * first_vector.y - e.y * first_vector.x
    squared_cross_product = cross_product * cross_product
    if squared_cross_product > (first_squared_length * e_squared_length
                                / squared_inv_epsilon):
        return 0, None, None
    s0 = first_vector.x * e.x + first_vector.y * e.y
    try:
        s0 /= first_squared_length
    except ArithmeticError:
        s0 = s1 = math.nan
    else:
        s1 = s0 + (first_vector.x * second_vector.x
                   + first_vector.y * second_vector.y) / first_squared_length
    s_min, s_max = min(s0, s1), max(s0, s1)
    (intersections_count, first_coefficient,
     second_coefficient) = _find_intersections(0, 1, s_min, s_max)
    first_intersection_point = second_intersection_point = None
    if intersections_count:
        first_intersection_point = Point(first_source.x
                                         + first_coefficient * first_vector.x,
                                         first_source.y
                                         + first_coefficient * first_vector.y)
        if first_intersection_point.distance_to(first_source) < threshold:
            first_intersection_point = first_source
        elif first_intersection_point.distance_to(first_target) < threshold:
            first_intersection_point = first_target
        elif first_intersection_point.distance_to(second_source) < threshold:
            first_intersection_point = second_source
        elif first_intersection_point.distance_to(second_target) < threshold:
            first_intersection_point = second_target
        if intersections_count > 1:
            second_intersection_point = Point(first_source.x
                                              + second_coefficient
                                              * first_vector.x,
                                              first_source.y
                                              + second_coefficient
                                              * first_vector.y)
    return (intersections_count, first_intersection_point,
            second_intersection_point)



import math

from reprit.base import generate_repr

from .bounding_box import BoundingBox
from .hints import Scalar


class Point:
    __slots__ = ('_x', '_y')

    def __init__(self, x: Scalar, y: Scalar) -> None:
        self._x = x
        self._y = y

    __repr__ = generate_repr(__init__)

    @property
    def x(self) -> Scalar:
        return self._x

    @property
    def y(self) -> Scalar:
        return self._y

    @property
    def bounding_box(self) -> BoundingBox:
        return BoundingBox(self._x, self._y, self._x, self._y)

    def __eq__(self, other: 'Point') -> bool:
        return (self._x == other._x and self._y == other._y
                if isinstance(other, Point)
                else NotImplemented)

    def distance_to(self, other: 'Point') -> Scalar:
        return math.sqrt((self._x - other._x) ** 2 + (self._y - other._y) ** 2)



from functools import reduce
from operator import add
from typing import (Iterator,
                    List)

from reprit.base import generate_repr

from .bounding_box import BoundingBox
from .hints import Scalar
from .point import Point


class Contour:
    __slots__ = ('_points', '_holes', '_is_external')

    def __init__(self, points: List[Point], holes: List[int], is_external: bool
                 ) -> None:
        self._points = points
        self._holes = holes
        self._is_external = is_external

    __repr__ = generate_repr(__init__)

    def __iter__(self) -> Iterator[Point]:
        return iter(self._points)

    @property
    def points(self) -> List[Point]:
        return self._points

    @property
    def holes(self) -> List[int]:
        return self._holes

    @property
    def is_external(self) -> bool:
        return self._is_external

    @property
    def is_counterclockwise(self) -> bool:
        points = self._points
        signed_area = sum(
                points[index].x * points[(index + 1) % len(points)].y
                - points[(index + 1) % len(points)].x * points[index].y
                for index in range(len(points)))
        return signed_area >= 0

    @property
    def is_clockwise(self) -> bool:
        return not self.is_counterclockwise

    @property
    def bounding_box(self) -> BoundingBox:
        if self._points:
            return reduce(add, [point.bounding_box for point in self._points])
        else:
            return BoundingBox(0, 0, 0, 0)

    def __eq__(self, other: 'Contour') -> bool:
        return (self._points == other._points
                and self._holes == other._holes
                and self._is_external is other._is_external
                if isinstance(other, Contour)
                else NotImplemented)

    def add(self, point: Point) -> None:
        self._points.append(point)

    def add_hole(self, hole: int) -> None:
        self._holes.append(hole)

    def clear_holes(self) -> None:
        self._holes.clear()

    def reverse(self) -> None:
        self._points = self._points[::-1]

    def set_clockwise(self) -> None:
        if self.is_counterclockwise:
            self.reverse()

    def set_counterclockwise(self) -> None:
        if self.is_clockwise:
            self.reverse()



from functools import reduce
from operator import add
from typing import (Iterator,
                    List)

from reprit.base import generate_repr

from .bounding_box import BoundingBox
from .contour import Contour


class Polygon:
    __slots__ = ('_contours',)

    def __init__(self, contours: List[Contour]) -> None:
        self._contours = contours

    __repr__ = generate_repr(__init__)

    @property
    def contours(self) -> List[Contour]:
        return self._contours

    def __eq__(self, other: 'Polygon') -> bool:
        return (self._contours == other._contours
                if isinstance(other, Polygon)
                else NotImplemented)

    def __iter__(self) -> Iterator[Contour]:
        return iter(self._contours)

    @property
    def bounding_box(self) -> BoundingBox:
        if self._contours:
            return reduce(add, [contour.bounding_box
                                for contour in self._contours])
        else:
            return BoundingBox(0, 0, 0, 0)

    def join(self, other: 'Polygon') -> None:
        contours_count = len(self._contours)
        self._contours.extend(Contour(contour.points,
                                      [hole + contours_count
                                       for hole in contour.holes],
                                      contour.is_external)
                              for contour in other._contours)



import enum
from functools import (partial,
                       partialmethod)
from operator import (attrgetter,
                      itemgetter)
from reprlib import recursive_repr
from typing import (Callable,
                    List,
                    Optional,
                    TypeVar,
                    Union)

from reprit.base import generate_repr

from .point import Point
from .segment import Segment
from .utilities import sign

ACYCLIC_INDEX = -1


class _EnumBase(enum.IntEnum):
    def __repr__(self) -> str:
        return self.__class__.__qualname__ + '.' + self.name


@enum.unique
class EdgeType(_EnumBase):
    NORMAL = 0
    NON_CONTRIBUTING = 1
    SAME_TRANSITION = 2
    DIFFERENT_TRANSITION = 3


@enum.unique
class OperationType(_EnumBase):
    INTERSECTION = 0
    UNION = 1
    DIFFERENCE = 2
    XOR = 3


@enum.unique
class PolygonType(_EnumBase):
    SUBJECT = 0
    CLIPPING = 1


Domain = TypeVar('Domain')


def _fill_chain(start: Domain, chain: List[Domain],
                to_next: Callable[[Domain], Optional[Domain]]) -> int:
    chain.append(start)
    cursor = to_next(start)
    while cursor is not None:
        try:
            cycle_index = next(index
                               for index, element in enumerate(chain)
                               if element is cursor)
        except StopIteration:
            chain.append(cursor)
            cursor = to_next(cursor)
        else:
            # last element points to already visited one
            # with this index
            return cycle_index
    # has no cycles
    return ACYCLIC_INDEX


SweepEventState = List[Union[bool, Point, Optional[list],
                             PolygonType, EdgeType]]


class SweepEvent:
    __slots__ = ('is_left', 'point', 'other_event', 'polygon_type',
                 'edge_type')
    OTHER_EVENT_STATE_INDEX = 2

    def __init__(self, is_left: bool, point: Point,
                 other_event: Optional['SweepEvent'],
                 polygon_type: PolygonType, edge_type: EdgeType) -> None:
        self.is_left = is_left
        self.point = point
        self.other_event = other_event
        self.polygon_type = polygon_type
        self.edge_type = edge_type

    def __getstate__(self) -> SweepEventState:
        chain = []  # type: List[SweepEvent]
        cycle_index = self._fill_sweep_events_chain(chain)
        states = [[sweep_event.is_left, sweep_event.point, None,
                   sweep_event.polygon_type, sweep_event.edge_type]
                  for sweep_event in chain]
        for index in range(len(states) - 1):
            states[index][self.OTHER_EVENT_STATE_INDEX] = states[index + 1]
        if cycle_index != ACYCLIC_INDEX:
            states[-1][self.OTHER_EVENT_STATE_INDEX] = states[cycle_index]
        return states[0]

    def __setstate__(self, state: SweepEventState) -> 'SweepEvent':
        (self.is_left, self.point, self.other_event,
         self.polygon_type, self.edge_type) = (state[0], state[1], None,
                                               state[3], state[4])
        chain = []  # type: List[SweepEventState]
        cycle_index = self._fill_states_chain(state, chain)
        sweep_events = [self] + [SweepEvent(state[0], state[1], None,
                                            state[3], state[4])
                                 for state in chain[1:]]
        for index in range(len(sweep_events) - 1):
            sweep_events[index].other_event = sweep_events[index + 1]
        if cycle_index != ACYCLIC_INDEX:
            sweep_events[-1].other_event = sweep_events[cycle_index]

    __repr__ = recursive_repr()(generate_repr(__init__))

    def __eq__(self, other: 'SweepEvent') -> bool:
        if self is other:
            return True

        def are_fields_equal(left: SweepEvent, right: SweepEvent) -> bool:
            return (left.is_left is right.is_left
                    and left.point == right.point
                    and left.polygon_type is right.polygon_type
                    and left.edge_type is right.edge_type)

        chain, other_chain = [], []
        return (are_fields_equal(self, other)
                and (self._fill_sweep_events_chain(chain)
                     == other._fill_sweep_events_chain(other_chain))
                and len(chain) == len(other_chain)
                and all(are_fields_equal(child, other_child)
                        for child, other_child in zip(chain[1:],
                                                      other_chain[1:]))
                if isinstance(other, SweepEvent)
                else NotImplemented)

    _fill_sweep_events_chain = partialmethod(_fill_chain,
                                             to_next=attrgetter('other_event'))
    _fill_states_chain = staticmethod(
            partial(_fill_chain,
                    to_next=itemgetter(OTHER_EVENT_STATE_INDEX)))

    @property
    def is_vertical(self) -> bool:
        self.validate()
        return self.point.x == self.other_event.point.x

    @property
    def segment(self) -> Segment:
        self.validate()
        return Segment(self.point, self.other_event.point)

    def is_above(self, point: Point) -> bool:
        return not self.is_below(point)

    def is_below(self, point: Point) -> bool:
        self.validate()
        return (sign(self.point, self.other_event.point, point) == 1
                if self.is_left
                else sign(self.other_event.point, self.point, point) == 1)

    def validate(self) -> None:
        if self.other_event is None:
            raise ValueError('No "other_event" found.')


class EventsQueueKey:
    __slots__ = ('_event',)

    def __init__(self, event: SweepEvent) -> None:
        self._event = event

    __repr__ = generate_repr(__init__)

    @property
    def event(self) -> SweepEvent:
        return self._event

    def __eq__(self, other: 'EventsQueueKey') -> bool:
        return (self._event == other._event
                if isinstance(other, EventsQueueKey)
                else NotImplemented)

    def __lt__(self, other: 'EventsQueueKey') -> bool:
        if not isinstance(other, EventsQueueKey):
            return NotImplemented
        if self._event.point.x != other._event.point.x:
            # different x-coordinate,
            # the event with lower x-coordinate is processed first
            return self._event.point.x > other._event.point.x
        if self._event.point.y != other._event.point.y:
            # different points, but same x-coordinate,
            # the event with lower y-coordinate is processed first
            return self._event.point.y > other._event.point.y
        if self._event.is_left is not other._event.is_left:
            # same point, but one is a left endpoint
            # and the other a right endpoint,
            # the right endpoint is processed first
            return self._event.is_left
        # same point, both events are left endpoints
        # or both are right endpoints
        if sign(self._event.point, self._event.other_event.point,
                other._event.other_event.point):  # not collinear
            # the event associate to the bottom segment is processed first
            return self._event.is_above(other._event.other_event.point)
        return self._event.polygon_type > other._event.polygon_type


class SweepLineKey:
    __slots__ = ('_event',)

    def __init__(self, event: SweepEvent) -> None:
        self._event = event

    __repr__ = generate_repr(__init__)

    @property
    def event(self) -> SweepEvent:
        return self._event

    def __eq__(self, other: 'SweepLineKey') -> bool:
        return (self._event == other._event
                if isinstance(other, SweepLineKey)
                else NotImplemented)

    def __lt__(self, other: 'SweepLineKey') -> bool:
        if not isinstance(other, SweepLineKey):
            return NotImplemented
        if self is other:
            return False
        if (sign(self.event.point, self.event.other_event.point,
                 other.event.point)
                or sign(self.event.point, self.event.other_event.point,
                        other.event.other_event.point)):
            # segments are not collinear
            if self.event.point == other.event.point:
                # same left endpoint, use the right endpoint to sort
                return self.event.is_below(other.event.other_event.point)
            # different left endpoint, use the left endpoint to sort
            elif self.event.point.x == other.event.point.x:
                return self.event.point.y < other.event.point.y
            elif EventsQueueKey(self.event) < EventsQueueKey(other.event):
                # has the line segment associated to `self` been inserted
                # into sweep line after the line segment associated to `other`?
                return other.event.is_above(self.event.point)
            # the line segment associated to `other` has been inserted
            # into sweep line after the line segment associated to `self`
            return self.event.is_below(other.event.point)
        # segments are collinear
        return (self.event.polygon_type < other.event.polygon_type
                if self.event.polygon_type is not other.event.polygon_type
                else EventsQueueKey(self.event) < EventsQueueKey(other.event))



from operator import attrgetter

from reprit.base import generate_repr

from .point import Point

points_key = attrgetter('x', 'y')


class Segment:
    __slots__ = ('source', 'target')

    def __init__(self, source: Point, target: Point):
        self.source = source
        self.target = target

    __repr__ = generate_repr(__init__)

    def __eq__(self, other: 'Segment') -> bool:
        return (self.source == other.source and self.target == other.target
                if isinstance(other, Segment)
                else NotImplemented)

    @property
    def max(self) -> Point:
        return max(self.source, self.target,
                   key=points_key)

    @property
    def min(self) -> Point:
        return min(self.source, self.target,
                   key=points_key)

    @property
    def is_degenerate(self) -> bool:
        return self.source == self.target

    @property
    def is_vertical(self) -> bool:
        return self.source.x == self.target.x

    @property
    def reversed(self) -> 'Segment':
        return Segment(self.target, self.source)



from reprit.base import generate_repr

from .hints import Scalar


class BoundingBox:
    __slots__ = ('_x_min', '_y_min', '_x_max', '_y_max')

    def __init__(self, x_min: Scalar, y_min: Scalar,
                 x_max: Scalar, y_max: Scalar) -> None:
        self._x_min = x_min
        self._y_min = y_min
        self._x_max = x_max
        self._y_max = y_max

    __repr__ = generate_repr(__init__)

    @property
    def x_min(self) -> Scalar:
        return self._x_min

    @property
    def y_min(self) -> Scalar:
        return self._y_min

    @property
    def x_max(self) -> Scalar:
        return self._x_max

    @property
    def y_max(self) -> Scalar:
        return self._y_max

    def __eq__(self, other: 'BoundingBox') -> bool:
        return (self._x_min == other._x_min
                and self._y_min == other._y_min
                and self._x_max == other._x_max
                and self._y_max == other._y_max
                if isinstance(other, BoundingBox)
                else NotImplemented)

    def __add__(self, other: 'BoundingBox') -> 'BoundingBox':
        return BoundingBox(min(self._x_min, other._x_min),
                           min(self._y_min, other._y_min),
                           max(self._x_max, other._x_max),
                           max(self._y_max, other._y_max))
